/* chmem - set total memory size for execution	Author: Andy Tanenbaum */

#include <errno.h>
#include <stdarg.h>
#include <stdlib.h>
#include <stdint.h>
#include <stdio.h>

#define MAX_8086     0x10000L	/* maximum allocation size for 8086 */
#define MAX_386	  0x7FFFFFFFL	/* etc */
#define MAX_68K   0x7FFFFFFFL
#define MAX_SPARC 0x20000000L	/* No more than 512MB on a SparcStation! */

struct mnx_hdr {
	uint8_t  a_magic[2];
	uint8_t  a_flags;
	uint8_t  a_cpu;
	uint8_t  a_hdrlen;
	uint8_t  a_unused;
	uint16_t a_version;
	uint32_t a_text;
	uint32_t a_data;
	uint32_t a_bss;
	uint32_t a_entry;
	uint32_t a_total;
	uint32_t a_syms;
};

#define A_MAGIC0      0x01
#define A_MAGIC1      0x03
#define BADMAG(X)     ((X).a_magic[0] != A_MAGIC0 ||(X).a_magic[1] != A_MAGIC1)

#define A_I8086 0x04    /* intel i8086/8088 */
#define A_M68K  0x0B    /* motorola m68000 */
#define A_I80386 0x10   /* intel i80386 */
#define A_SPARC 0x17    /* Sun SPARC */

#define A_UZP   0x01    /* unmapped zero page (pages) */
#define A_EXEC  0x10    /* executable */
#define A_SEP   0x20    /* separate I/D */

static uint32_t getuint32(FILE *);
static uint16_t getuint16(FILE *);
static uint8_t getuint8(FILE *);
static void putuint32(uint32_t, FILE *);
static void putuint16(uint16_t, FILE *);
static void putuint8(uint8_t, FILE *);

char *progname;

static void fatal(const char *, ...);
void error(char *s1, char *s2);
void usage(void);

int main(int argc, char *argv[])
{
/* The 8088 architecture does not make it possible to catch stacks that grow
 * big.  The only way to deal with this problem is to let the stack grow down
 * towards the data segment and the data segment grow up towards the stack.
 * Normally, a total of 64K is allocated for the two of them, but if the
 * programmer knows that a smaller amount is sufficient, he can change it
 * using chmem.
 *
 * chmem =4096 prog  sets the total space for stack + data growth to 4096
 * chmem +200  prog  increments the total space for stack + data growth by 200
 */

	char *p;
	int separate;
	long lsize, olddynam, newdynam, newtot, overflow;
	FILE *file;
	struct mnx_hdr mexec;
	char cpu;
	long max;

	newdynam = 0;	/* silence gcc -Wall */

	progname = argv[0];
	if (argc < 3)
		usage();
	p = argv[1];
	if (*p != '=' && *p != '+' && *p != '-')
		usage();
	lsize = atol(p + 1);

	if (lsize < 0) {
		error(p + 1, "is negative");
		exit(EXIT_FAILURE);
	}
	argc -= 1;
	argv += 1;

	while (--argc) {
		++argv;
		if ((file = fopen(*argv, "r+")) == NULL) {
			error("can't open", *argv);
			continue;
		}

		mexec.a_magic[0] = getuint8(file);
		mexec.a_magic[1] = getuint8(file);
		mexec.a_flags = getuint8(file);
		mexec.a_cpu = getuint8(file);
		mexec.a_hdrlen = getuint8(file);
		mexec.a_unused = getuint8(file);
		mexec.a_version = getuint16(file);
		mexec.a_text = getuint32(file);
		mexec.a_data = getuint32(file);
		mexec.a_bss = getuint32(file);
		mexec.a_entry = getuint32(file);
		mexec.a_total = getuint32(file);
		mexec.a_syms = getuint32(file);

		if (BADMAG(mexec)) {
			error(*argv, "is not executable");
			continue;
		}
		separate = (mexec.a_flags & A_SEP ? 1 : 0);
		cpu = mexec.a_cpu;

		switch (cpu) {
		    case A_I8086:	max = MAX_8086;	break;
		    case A_I80386:	max = MAX_386;	break;
		    case A_M68K:	max = MAX_68K;	break;
		    case A_SPARC:	max = MAX_SPARC;	break;
		    default:
			error("bad CPU type in", *argv);
			continue;
		}

		if (lsize > max) {
			error("size is too large for", *argv);
			continue;
		}
		olddynam = mexec.a_total - mexec.a_data - mexec.a_bss;
		if (separate == 0)
			olddynam -= mexec.a_text;

		if (*p == '=')
			newdynam = lsize;
		else if (*p == '+')
			newdynam = olddynam + lsize;
		else if (*p == '-')
			newdynam = olddynam - lsize;

		newtot = mexec.a_data + mexec.a_bss + newdynam;
		if (separate == 0)
			newtot += mexec.a_text;
		overflow = (newtot > max ? newtot - max : 0);
		newdynam -= overflow;
		newtot -= overflow;
		mexec.a_total = newtot;
		rewind(file);
		putuint8(mexec.a_magic[0], file);
		putuint8(mexec.a_magic[1], file);
		putuint8(mexec.a_flags, file);
		putuint8(mexec.a_cpu, file);
		putuint8(mexec.a_hdrlen, file);
		putuint8(mexec.a_unused, file);
		putuint16(mexec.a_version, file);
		putuint32(mexec.a_text, file);
		putuint32(mexec.a_data, file);
		putuint32(mexec.a_bss, file);
		putuint32(mexec.a_entry, file);
		putuint32(mexec.a_total, file);
		putuint32(mexec.a_syms, file);
	
		printf("%s: Stack+malloc area changed from %ld to %ld bytes.\n",
		       *argv, olddynam, newdynam);
		fclose(file);
	}
	return(0);
}

void error(s1, s2)
char *s1;
char *s2;
{
  fprintf(stderr, "%s: %s ", progname, s1);
  if (errno != 0)
	perror(s2);
  else
	fprintf(stderr, "%s\n", s2);
  errno = 0;
}

void usage()
{
  fprintf(stderr, "Usage: %s {=+-} amount file\n", progname);
  exit(1);
}

static uint32_t
getuint32(FILE *input)
{
	uint32_t result;

	result = getuint8(input);
	result = result << 8 | getuint8(input);
	result = result << 8 | getuint8(input);
	result = result << 8 | getuint8(input);

	return result;
}

static uint16_t
getuint16(FILE *input)
{
	uint16_t result;

	result = getuint8(input);
	result = result << 8 | getuint8(input);

	return result;
}

static uint8_t
getuint8(FILE *input)
{
	int result;

	if ((result = getc(input)) == EOF)
		fatal("read error\n");

	return result;
}

static void
putuint32(uint32_t value, FILE *output)
{
	putuint8(value >> 24, output);
	putuint8(value >> 16, output);
	putuint8(value >>  8, output);
	putuint8(value      , output);

	return;
}

static void
putuint16(uint16_t value, FILE *output)
{
	putuint8(value >>  8, output);
	putuint8(value      , output);

	return;
}

static void
putuint8(uint8_t value, FILE *output)
{
	if (putc(value, output) == EOF)
		fatal("write error\n");

	return;
}

static void
fatal(const char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);
	fprintf(stderr, "%s: ", progname);
	vfprintf(stderr, fmt, ap);
	va_end(ap);

	exit(EXIT_FAILURE);
}
