/*-
 * Copyright (c) 1988-1992, 2005, 2009-2015, 2021, 2023
 * 	Hans Ottevanger.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * cv: converts executables from GNU M68K a.out format to MINIX-ST format.
 */

#include <sys/stat.h>

#include <stdarg.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/*
 * Our rendition of the GNU a.out header. For a description of the original,
 * see e.g. a.out(5) on a BSD system.
 * We use uint32_t to achieve wordsize independence.
 */

struct gnu_hdr {
	uint32_t a_magic;
	uint32_t a_text;
	uint32_t a_data;
	uint32_t a_bss;
	uint32_t a_syms;
	uint32_t a_entry;
	uint32_t a_trsize;
	uint32_t a_drsize;
};

/*
 * We define only those macros that we actually use. The original has far more
 * of them.
 */

#define GNU_MAGIC(hdr)		((hdr).a_magic & 0xffff)
#define GNU_MACHTYPE(hdr)	(((hdr).a_magic >> 16) & 0xff)

#define GNU_OMAGIC	0x107	/* old impure format */
#define GNU_M68020	2	/* generated by gas, even with -m68000 */

/*
 * In the original a.out.h, bitfields (endian dependent!) are used instead of a
 * single word r_info. We define macros only for the bits we actually expect to
 * use.
 */

struct gnu_reloc {
	uint32_t r_address;
	uint32_t r_info;
};

#define GNU_RELINFO(reloc)	((reloc).r_info & 0xff)
#define GNU_RELSYM(reloc)	((reloc).r_info >> 8)

#define GNU_RELEXT	0x10
#define GNU_REL2BYTE	0x20
#define GNU_REL4BYTE	0x40
#define GNU_RELPCREL	0x80

/*
 * Slighty simplified definition of a namelist entry. The original defines a
 * union containing a char *, which we cannot use, since that may be 64 bits
 * wide on some platforms.
 */

struct gnu_nlist {
	uint32_t n_strx;
	uint8_t  n_type;
	uint8_t  n_other;
	uint16_t n_desc;
	uint32_t n_value;
};

#define GNU_UNDF	0x00
#define GNU_EXT		0x01
#define GNU_ABS		0x02
#define GNU_TEXT	0x04
#define GNU_DATA	0x06
#define GNU_BSS		0x08
#define GNU_WEAK_UNDF	0x0c
#define GNU_WEAK	0x0e
#define GNU_WEAK_OBJ	0x10
#define GNU_TYPE	0x1e
#define GNU_STAB	0xe0

/*
 * Our version of the Minix header and namelist entry and corresponding macros.
 */

struct mnx_hdr {
	uint8_t  a_magic[2];
	uint8_t  a_flags;
	uint8_t  a_cpu;
	uint8_t  a_hdrlen;
	uint8_t  a_unused;
	uint16_t a_version;
	uint32_t a_text;
	uint32_t a_data;
	uint32_t a_bss;
	uint32_t a_entry;
	uint32_t a_total;
	uint32_t a_syms;
};

#define MNX_MAGIC0	0x01
#define MNX_MAGIC1	0x03
#define MNX_EXEC	0x10
#define MNX_M68K	0x0b

#define MNX_NSIZE	8

struct mnx_nlist {
	char     n_name[MNX_NSIZE];
	uint32_t n_value;
	uint8_t  n_sclass;
	uint8_t  n_numaux;
	uint16_t n_type;
};

#define MNX_UNDF	0x00
#define MNX_ABS		0x01
#define MNX_TEXT	0x02
#define MNX_DATA	0x03
#define MNX_BSS		0x04
#define MNX_EXT		0x10

/* Forward declarations. */

static void betohu32(uint32_t *);
static void betohu16(uint16_t *);
static void htobeu32(uint32_t *);
static void htobeu16(uint16_t *);
static int putbeu32(uint32_t, FILE *);
static int putu8(uint8_t, FILE *);
static int cmpu32(const void *, const void *);
static void setexec(const char *);
static void fatal(const char *, ...);
static void usage(void);
static void *xmalloc(size_t);

static char *progname;

int
main(int argc, char *argv[])
{
	char *inputfilename, *outputfilename;
	FILE *input, *output;
	struct gnu_hdr gexec;
	uint8_t *textsegment, *datasegment;
	uint32_t ntrel, ndrel, nrel;
	uint32_t nundefs, nsyms, nskipped;
	struct gnu_reloc *greloc, *grelp;
	struct gnu_nlist *gsymtab, *gsymp;
	char *strings;
	struct mnx_hdr mexec;
	uint32_t *mreloc, *mrelp;
	struct mnx_nlist *msymtab, *msymp;

	progname = argv[0];

	inputfilename = NULL;
	outputfilename = NULL;

	/* Parse arguments and options, and deal with them. */

	for (int i = 1; i < argc; i++) {
		if (argv[i][0] == '-')
			switch (argv[i][1]) {
			case 'h':
				usage();
				break;
			case 'o':
				if (argv[i][2] != '\0')
					outputfilename = argv[i] + 2;
				else if (++i < argc)
					outputfilename = argv[i];
				else {
					fprintf(stderr, "illegal argument\n");
					usage();
				}
				break;
			default:
				fprintf(stderr, "illegal option\n");
				usage();
			}
		else
			inputfilename = argv[i];
	}

	if (!outputfilename)
		outputfilename = "a.out";

	if (!inputfilename)
		input = stdin;
	else {
		if (strcmp(inputfilename, outputfilename) == 0)
			fatal("won't overwrite input file\n");
		if ((input = fopen(inputfilename, "r")) == NULL)
			fatal("can't open input file %s\n", inputfilename);
	}

	/* Read and validate the GNU header. */

	if (fread(&gexec, sizeof(gexec), 1, input) != 1)
		fatal("can't read header\n");
	betohu32(&gexec.a_magic);
	betohu32(&gexec.a_text);
	betohu32(&gexec.a_data);
	betohu32(&gexec.a_bss);
	betohu32(&gexec.a_syms);
	betohu32(&gexec.a_entry);
	betohu32(&gexec.a_trsize);
	betohu32(&gexec.a_drsize);

	if ((GNU_MAGIC(gexec) != GNU_OMAGIC) ||
	    (GNU_MACHTYPE(gexec) != GNU_M68020))
		fatal("unexpected magic 0x%lx\n", gexec.a_magic);

	if (gexec.a_entry != 0)
		fatal("can't handle non-zero entry point\n");

	/*
	 * Read text and data segments and store them until we have generated
	 * and written the Minix header.
	 */

	textsegment = xmalloc(gexec.a_text);
	datasegment = xmalloc(gexec.a_data);

	if ((gexec.a_text != 0) &&
	    (fread(textsegment, gexec.a_text, 1, input) != 1))
		fatal("can't read text segment\n");
	if ((gexec.a_data != 0) &&
	    (fread(datasegment, gexec.a_data, 1, input) != 1))
		fatal("can't read data segment\n");

	/*
	 * Read the text and data relocation entries into a single array.
	 * Postpone validating them until later, when we are sure we don't have
	 * any undefined symbols. Doing otherwise could trigger a spurious
	 * fatal "unusable relocation information" message, before we even see
	 * the list of undefined symbols.
	 */

	ntrel = gexec.a_trsize / sizeof(struct gnu_reloc);
	ndrel = gexec.a_drsize / sizeof(struct gnu_reloc);
	nrel = ntrel + ndrel;

	greloc = xmalloc(nrel*sizeof(struct gnu_reloc));

	for (grelp = greloc; grelp < greloc + ntrel; grelp++) {
		if (fread(grelp, sizeof(struct gnu_reloc), 1, input) != 1)
		    fatal("can't read text relocation info\n");
		betohu32(&grelp->r_address);
		betohu32(&grelp->r_info);
	}
	for (; grelp < greloc + nrel; grelp++) {
		if (fread(grelp, sizeof(struct gnu_reloc), 1, input) != 1)
		    fatal("can't read data relocation info\n");
		betohu32(&grelp->r_address);
		betohu32(&grelp->r_info);
		grelp->r_address += gexec.a_text;
	}

	/* Read the symbol table and the string table. */

	gsymtab = xmalloc(gexec.a_syms);
	nsyms = gexec.a_syms / sizeof(struct gnu_nlist);

	for (gsymp = gsymtab; gsymp < gsymtab + nsyms; gsymp++) {
		if (fread(gsymp, sizeof(struct gnu_nlist), 1, input) != 1)
		    fatal("can't read symbol table\n");
		betohu32(&gsymp->n_strx);
		betohu16(&gsymp->n_desc);
		betohu32(&gsymp->n_value);
	}

	if (nsyms > 0) {
		uint32_t strsize;

		if (fread(&strsize, sizeof(strsize), 1, input) != 1)
		    fatal("can't read size of strings\n");
		betohu32(&strsize);

		strings = xmalloc(strsize);

		if (fread(strings + sizeof(strsize),
		    strsize - sizeof(strsize), 1, input) != 1)
			fatal("can't read strings\n");
	} else
		strings = NULL;

	/*
	 * Convert the GNU symbol table information into the Minix symbol table
	 * format. We keep track of the number of entries we need to skip due to
	 * being of a type unknown to the Minix format. Other symbol types are
	 * converted to their corresponding type. Symbol names are truncated to
	 * MNX_NSIZE characters (currently 8). Because the linker (as used with
	 * the present set of options) may silently leave symbols undefined, we
	 * have to check for undefined symbols and report them here.
	 */

	msymtab = xmalloc(nsyms * sizeof(struct mnx_nlist));
	msymp = msymtab;
	nundefs = 0;
	nskipped = 0;

	for (gsymp = gsymtab; gsymp < gsymtab + nsyms; gsymp++) {
		if (gsymp->n_type & GNU_STAB) {
			nskipped++;
			continue;
		}

		if ((gsymp->n_type & GNU_TYPE) == GNU_WEAK_UNDF) {
			nskipped++;
			continue;
		}

		if (gsymp->n_type & GNU_EXT)
			msymp->n_sclass = MNX_EXT;
		else
			msymp->n_sclass = 0;
		
		if (gsymp->n_strx == 0)
			fatal("unnamed symbol table entry\n");
		else
			strncpy(msymp->n_name,
			    strings + gsymp->n_strx, MNX_NSIZE);

		msymp->n_value = gsymp->n_value;

		switch (gsymp->n_type & GNU_TYPE) {
		case GNU_UNDF:
			if (nundefs == 0)
				fprintf(stderr, "Undefined symbols:\n");
			nundefs++;
			fprintf(stderr, "\t%s\n", strings + gsymp->n_strx);
			msymp->n_sclass |= MNX_UNDF;
			break;
		case GNU_ABS:
			msymp->n_sclass |= MNX_ABS;
			break;
		case GNU_TEXT:
			msymp->n_sclass |= MNX_TEXT;
			break;
		case GNU_DATA:
			msymp->n_sclass |= MNX_DATA;
			break;
		case GNU_BSS:
			msymp->n_sclass |= MNX_BSS;
			break;
		case GNU_WEAK:
		case GNU_WEAK_OBJ:
			if (gsymp->n_value < gexec.a_text)
				msymp->n_sclass |= MNX_TEXT;
			else if (gsymp->n_value < gexec.a_text + gexec.a_data)
				msymp->n_sclass |= MNX_DATA;
			else
				msymp->n_sclass |= MNX_BSS;
			break;
		default:
			fatal("unknown symbol type: %s = %x\n",
			    strings + gsymp->n_strx,
			    gsymp->n_type & GNU_TYPE);
			break;
		}

		msymp->n_numaux = 0;
		msymp->n_type = 0;
		msymp++;
	}

	if (nundefs != 0)
		exit(EXIT_FAILURE);
	nsyms -= nskipped;

	/*
	 * Check the relocation information and convert it into a single array
	 * of address values to be relocated on run time. The GNU relocation
	 * info could potentially assume far more different values than we will
	 * ever encounter: we check only for the subset that actually occurs.
	 * Entries that need no action are skipped and accounted for. We finally
	 * make sure the array is sorted, as required for further processing.
	 */

	mreloc = xmalloc(nrel * sizeof(uint32_t));
	mrelp = mreloc;
	nskipped = 0;

	for (grelp = greloc; grelp < greloc + nrel; grelp++) {
		if (grelp->r_address % 2U)
			fatal("odd relocation address\n");
		switch (GNU_RELINFO(*grelp)) {
		case GNU_REL4BYTE:
			*mrelp++ = grelp->r_address;
			break;
		case GNU_REL2BYTE|GNU_RELPCREL:
		case GNU_REL4BYTE|GNU_RELPCREL:
			nskipped++;
			break;
		case GNU_REL4BYTE|GNU_RELEXT:
			if ((gsymtab[GNU_RELSYM(*grelp)].n_type & GNU_TYPE)
			    != GNU_WEAK_UNDF)
				fatal("rogue external relocation\n");
			nskipped++;
			break;
		default:
			fatal("unusable relocation information: %x\n",
				GNU_RELINFO(*grelp));
			break;
		}
	}

	nrel -= nskipped;

	qsort(mreloc, nrel, sizeof(uint32_t), cmpu32);

	/*
	 * At last we have all the info to fill the Minix header.
	 * The total size of the process on runtime is rounded up to the next
	 * multiple of 64K. That size may be changed later with the chmem
	 * command.
	 */

	mexec.a_magic[0] = MNX_MAGIC0;
	mexec.a_magic[1] = MNX_MAGIC1;
	mexec.a_flags = MNX_EXEC;
	mexec.a_cpu = MNX_M68K;
	mexec.a_hdrlen = sizeof(mexec);
	mexec.a_unused = 0;
	mexec.a_version = 0;
	mexec.a_text = gexec.a_text;
	mexec.a_data = gexec.a_data;
	mexec.a_bss = gexec.a_bss;
	mexec.a_entry = gexec.a_entry;
	mexec.a_total =
	    (gexec.a_text + gexec.a_data + gexec.a_bss + 0x10000) & ~0xffffL;
	mexec.a_syms = nsyms * sizeof(struct mnx_nlist);

	/*
	 * We are finally ready to write the Minix executable.
	 * Nothing can go wrong from here on, except for output errors.
	 */

	if ((output = fopen(outputfilename, "w")) == NULL)
		fatal("can't open output file %s\n", outputfilename);

	/* Convert and write the header */

	htobeu16(&mexec.a_version);
	htobeu32(&mexec.a_text);
	htobeu32(&mexec.a_data);
	htobeu32(&mexec.a_bss);
	htobeu32(&mexec.a_entry);
	htobeu32(&mexec.a_total);
	htobeu32(&mexec.a_syms);
	if (fwrite(&mexec, sizeof(struct mnx_hdr), 1, output) != 1)
	    fatal("can't write minix header\n");
	
	/* Write text and data segments */

	if ((gexec.a_text != 0) &&
	    (fwrite(textsegment, gexec.a_text, 1, output) != 1))
		fatal("can't write text segment\n");
	if ((gexec.a_data != 0) &&
	    (fwrite(datasegment, gexec.a_data, 1, output) != 1))
		fatal("can't write data segment\n");

	/* Convert and write the symbol table */

	for (msymp = msymtab; msymp < msymtab + nsyms; msymp++) {
		htobeu32(&msymp->n_value);
		htobeu16(&msymp->n_type);
		if (fwrite(msymp, sizeof(struct mnx_nlist), 1, output) != 1)
		    fatal("can't write symbol table\n");
	}

	/*
	 * Write the relocation information in the GEMDOS format. See the
	 * comment for relocate() in src/mm/exec.c for a description.
	 */

	if (nrel == 0) {
		if (putbeu32(0, output) != 0)
		    fatal("can't write relocation info\n");
	} else {
		uint32_t base, jump;

		mrelp = mreloc;
		base = *mrelp;
		if (putbeu32(base, output) != 0)
		    fatal("can't write relocation info\n");
		mrelp++;
		while (mrelp < mreloc + nrel) {
			jump = *mrelp - base;
			while (jump > 254) {
				if (putu8(1, output) != 0)
				    fatal("can't write relocation info\n");
				jump -= 254;
			}
			if (putu8(jump, output) != 0)
			    fatal("can't write relocation info\n");
			base = *mrelp;
			mrelp++;
		}

		if (putu8(0, output) != 0)
		    fatal("can't write relocation info\n");
	}

	/* Clean up after ourselves */

	free(textsegment);
	free(datasegment);
	free(greloc);
	free(gsymtab);
	free(strings);
	free(msymtab);
	free(mreloc);

	if (input != stdin)
		fclose(input);
	fclose(output);

	setexec(outputfilename);

	exit(EXIT_SUCCESS);
}

/*
 * Make file filename executable, taking into account the current umask.
 * Note: umask() cannot fail.
 */

static void
setexec(const char *filename)
{
	mode_t mask;

	umask(mask = umask(0));
	if (chmod(filename, (S_IRWXU|S_IRWXG|S_IRWXO) & ~mask) < 0)
		fatal("can't change permissions of file %s\n", filename);
}

#ifdef __m68k__	/* Actually: big-endian */

/*
 * Functions to convert from big-endian to host byte order. These are no-ops
 * in case of a big-endian host.
 */

static void
betohu32(uint32_t *p)
{
	return;
}

static void
betohu16(uint16_t *p)
{
	return;
}

static void
htobeu32(uint32_t *p)
{
	return;
}

static void
htobeu16(uint16_t *p)
{
	return;
}

#else

/*
 * Functions to convert from big-endian to host byte order. These are for
 * little-endian hosts.
 */

static void
swapu32(uint32_t *p)
{
	uint32_t value, result;

	value = *p;

	result = value & 0xff;
	value >>= 8;
	result = (result << 8) | (value & 0xff);
	value >>= 8;
	result = (result << 8) | (value & 0xff);
	value >>= 8;
	result = (result << 8) | (value & 0xff);

	*p = result;

	return;
}

static void
swapu16(uint16_t *p)
{
	*p = *p << 8 | *p >> 8;

	return;
}

static void
betohu32(uint32_t *p)
{
	swapu32(p);

	return;
}

static void
betohu16(uint16_t *p)
{
	swapu16(p);

	return;
}

static void
htobeu32(uint32_t *p)
{
	swapu32(p);

	return;
}

static void
htobeu16(uint16_t *p)
{
	swapu16(p);

	return;
}

#endif

/*
 * Write an unsigned 32 bits integer to file output, converting it from host
 * to big-endian byte order on the fly.
 */

static int
putbeu32(uint32_t value, FILE *output)
{
	htobeu32(&value);

	if (fwrite(&value, sizeof(uint32_t), 1, output) != 1)
		return EOF;
	else
		return 0;
}

/*
 * Write an unsigned 8 bits integer to file output.
 */

static int
putu8(uint8_t value, FILE *output)
{
	if (putc(value, output) == EOF)
		return EOF;
	else
		return 0;
}

/* Compare two unsigned 32 bits integers. Only used to sort relocation info. */

static int
cmpu32(const void *p, const void *q)
{
	if (*(uint32_t *)p > *(uint32_t *)q)
		return 1;
	else if (*(uint32_t *)p < *(uint32_t *)q)
		return -1;
	else
		return 0;
}

/* Exit after printing a formatted message. */

static void
fatal(const char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);
	fprintf(stderr, "%s: ", progname);
	vfprintf(stderr, fmt, ap);
	va_end(ap);

	exit(EXIT_FAILURE);
}

static void
usage(void)
{
	fprintf(stderr,
	    "Usage: %s [-h] [gnuexecutable] [-o minixexecutable]\n", progname);

	exit(EXIT_FAILURE);
}

static void *
xmalloc(size_t n)
{
	void *mem;

	if ((mem = malloc(n)) == NULL)
		fatal("out of memory\n");

	return mem;
}
