/* nm - print name list.		Author: Dick van Veen */

/* Dick van Veen: veench@cs.vu.nl */

#include <sys/types.h>
#include <fcntl.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

/* Read the name list in memory, sort it, and print it.  */

/* Nm [-gnopru] [file] ...
 *
 * flags:
 *	-d	address in decimal
 *	-g	print only external symbols.
 *	-n	sort numerically rather than alphabetically.
 *	-o	prepend file name to each line rather than only once.
 *	-p	don't sort, pint n symbol-table order.
 *	-r	sort in reverse order.
 *	-u	print only undefined symbols.
 *
 *	-	when no file name is present, a.out is assumed.
 *
 *	NOTE:	no archives are supported because assembly files don't
 *		have symbol tables.
 *
 */

struct mnx_hdr {
	uint8_t  a_magic[2];
	uint8_t  a_flags;
	uint8_t  a_cpu;
	uint8_t  a_hdrlen;
	uint8_t  a_unused;
	uint16_t a_version;
	uint32_t a_text;
	uint32_t a_data;
	uint32_t a_bss;
	uint32_t a_entry;
	uint32_t a_total;
	uint32_t a_syms;
};

struct mnx_nlist {
	char     n_name[8];
	uint32_t n_value;
	uint8_t  n_sclass;
	uint8_t  n_numaux;
	uint16_t n_type;
};

#define A_MAGIC0      0x01
#define A_MAGIC1      0x03
#define BADMAG(X)     ((X).a_magic[0] != A_MAGIC0 ||(X).a_magic[1] != A_MAGIC1)

#define A_UZP   0x01    /* unmapped zero page (pages) */
#define A_EXEC  0x10    /* executable */
#define A_SEP   0x20    /* separate I/D */

#define A_TEXTPOS(X)    ((uint32_t)(X).a_hdrlen)
#define A_DATAPOS(X)    (A_TEXTPOS(X) + (X).a_text)
#define A_TRELPOS(X)    (A_DATAPOS(X) + (X).a_data)
#define A_SYMPOS(X)     (A_TRELPOS(X))

/* Low bits of storage class (section). */
#define N_SECT            07    /* section mask */
#define N_UNDF            00    /* undefined */
#define N_ABS             01    /* absolute */
#define N_TEXT            02    /* text */
#define N_DATA            03    /* data */
#define N_BSS             04    /* bss */
#define N_COMM            05    /* (common) */

/* High bits of storage class. */
#define N_CLASS         0370    /* storage class mask */
#define C_EXT           0020    /* external symbol */
#define C_STAT          0030    /* static */

static uint32_t getuint32(FILE *);
static uint16_t getuint16(FILE *);
static uint8_t getuint8(FILE *);
static void fatal(const char *fmt, ...);

#define A_OUT		"a.out"

int d_flag;
int g_flag;
int n_flag;
int o_flag;
int p_flag;
int r_flag;
int u_flag;

struct mnx_hdr header;		/* header of a.out file */
int stbl_elems;			/* #elements in symbol table */

int nm_sort(const void *tmp_stbl1, const void *tmp_stbl2);
void nm(const char *filename);
int read_header(FILE *file);
void nm_print(const char *filename, struct mnx_nlist *stbl);

char *progname;

int main(int argc, char *argv[])
{
	progname = argv[0];

	argv++;
	while (*argv != 0 && **argv == '-') {
		*argv += 1;
		while (**argv != '\0') {
			switch (**argv) {
			    case 'd':	d_flag = 1;	break;
			    case 'g':	g_flag = 1;	break;
			    case 'n':	n_flag = 1;	break;
			    case 'o':	o_flag = 1;	break;
			    case 'p':	p_flag = 1;	break;
			    case 'r':	r_flag = 1;	break;
			    case 'u':	u_flag = 1;	break;
			    default:
				fprintf(stderr, "illegal flag: -%c\n", **argv);
				exit(-1);
			}
			*argv += 1;
		}
		argv++;
 	}	
	if (*argv == 0)
		nm(A_OUT);
	else
		while (*argv != 0) {
			nm(*argv);
			argv++;
		}

	return 0;
}

int nm_sort(const void *tmp_stbl1, const void *tmp_stbl2)
{

	struct mnx_nlist *stbl1 = (struct mnx_nlist *)tmp_stbl1;
 	struct mnx_nlist *stbl2 = (struct mnx_nlist *)tmp_stbl2;
	int cmp;

	if (n_flag) {			/* sort numerically */
		if ((stbl1->n_sclass & N_SECT) <
		    (stbl2->n_sclass & N_SECT))
			cmp = -1;
		else if ((stbl1->n_sclass & N_SECT) >
			 (stbl2->n_sclass & N_SECT))
			cmp = 1;
		else if (stbl1->n_value < stbl2->n_value)
			cmp = -1;
		else if (stbl1->n_value > stbl2->n_value)
			cmp = 1;
		else
			cmp = strncmp(stbl1->n_name, stbl2->n_name, (size_t)8);
		} else {
		cmp = strncmp(stbl1->n_name, stbl2->n_name, (size_t)8);
		if (cmp == 0) {
			if (stbl1->n_value < stbl2->n_value)
				cmp = -1;
			else if (stbl1->n_value > stbl2->n_value)
				cmp = 1;
		}
	}

	if (r_flag) cmp = -cmp;	/* reverse sort */

	return cmp;
}

void nm(const char *filename)
{
	struct mnx_nlist *stbl, *p, *last;
	FILE *file;

	if ((file = fopen(filename, "r")) == NULL) {
		fprintf(stderr, "can't open %s\n", filename);
		return;
	}
	if (read_header(file)) {
		fprintf(stderr, "%s: no executable file\n", filename);
		fclose(file);
		return;
	}
	if (header.a_syms == 0) {
		fclose(file);
		return;
	}
	stbl = (struct mnx_nlist *) malloc((size_t) header.a_syms);
	if (stbl == NULL) {
		fprintf(stderr, "%s: can't allocate symbol table\n", filename);
		fclose(file);
		return;
	}
	stbl_elems = (int) header.a_syms / sizeof(struct mnx_nlist);
	for (p = stbl, last = &stbl[stbl_elems]; p < last; p++) {
		fread(p->n_name, 8, 1, file);
		p->n_value = getuint32(file);
		p->n_sclass = getuint8(file);
		p->n_numaux = getuint8(file);
		p->n_type = getuint16(file);
	}
	if (!p_flag) qsort(stbl, (size_t)stbl_elems, sizeof(struct mnx_nlist), nm_sort);
	nm_print(filename, stbl);
	free(stbl);
	fclose(file);

	return;
}

int read_header(FILE *file)
{
	header.a_magic[0] = getuint8(file);
	header.a_magic[1] = getuint8(file);
	header.a_flags = getuint8(file);
	header.a_cpu = getuint8(file);
	header.a_hdrlen = getuint8(file);
	header.a_unused = getuint8(file);
	header.a_version = getuint16(file);
	header.a_text = getuint32(file);
	header.a_data = getuint32(file);
	header.a_bss = getuint32(file);
	header.a_entry = getuint32(file);
	header.a_total = getuint32(file);
	header.a_syms = getuint32(file);

	if (BADMAG(header))
		return 1;

	fseek(file, A_SYMPOS(header), SEEK_SET);

	return 0;

}

void nm_print(const char *filename, struct mnx_nlist *stbl)
{
	struct mnx_nlist *last;
	char name[9];
	int n_sclass;
	char type;

	name[8] = '\0';
	if (!o_flag) printf("%s:\n", filename);
	for (last = &stbl[stbl_elems]; stbl != last; stbl++) {
		if (g_flag && (stbl->n_sclass & N_CLASS) != C_EXT) continue;
		if (u_flag && (stbl->n_sclass & N_SECT) != N_UNDF) continue;

		n_sclass = stbl->n_sclass & N_SECT;
		if (n_sclass == N_ABS)
			type = 'a';
		else if (n_sclass == N_TEXT)
			type = 't';
		else if (n_sclass == N_DATA)
			type = 'd';
		else if (n_sclass == N_BSS)
			type = 'b';
		else
			type = 'u';
		if ((stbl->n_sclass & N_CLASS) == C_EXT) type += 'A' - 'a';
		strncpy(name, stbl->n_name, (size_t)8);
		if (d_flag) {
			/* Offsets in decimal. */
			if (o_flag)
			       printf("%s:%08u %c %s\n",filename,stbl->n_value,type,name);
			else
			       printf("%08u %c %s\n", stbl->n_value, type, name);
		} else {
			/* Offsets in hex. */
			if (o_flag)
			       printf("%s:%08x %c %s\n",filename,stbl->n_value,type,name);
			else
			       printf("%08x %c %s\n", stbl->n_value, type, name);
		}
	}
}

static uint32_t
getuint32(FILE *input)
{
	uint32_t result;

	result = getuint8(input);
	result = result << 8 | getuint8(input);
	result = result << 8 | getuint8(input);
	result = result << 8 | getuint8(input);

	return result;
}

static uint16_t
getuint16(FILE *input)
{
	uint16_t result;

	result = getuint8(input);
	result = result << 8 | getuint8(input);

	return result;
}

static uint8_t
getuint8(FILE *input)
{
	int result;

	if ((result = getc(input)) == EOF)
		fatal("read error\n");

	return result;
}

static void
fatal(const char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);
	fprintf(stderr, "%s: ", progname);
	vfprintf(stderr, fmt, ap);
	va_end(ap);

	exit(EXIT_FAILURE);
}
