/*-
 * Copyright (c) 2014 Hans Ottevanger
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * mstrip: discard symbols from MINIX-ST executables.
 *
 */

#include <sys/stat.h>

#include <limits.h>
#include <stdarg.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

struct mnx_hdr {
	uint8_t  a_magic[2];
	uint8_t  a_flags;
	uint8_t  a_cpu;
	uint8_t  a_hdrlen;
	uint8_t  a_unused;
	uint16_t a_version;
	uint32_t a_text;
	uint32_t a_data;
	uint32_t a_bss;
	uint32_t a_entry;
	uint32_t a_total;
	uint32_t a_syms;
};

#define MNX_MAGIC0	0x01
#define MNX_MAGIC1	0x03
#define MNX_M68K	0x0b

static void strip(const char *);
static uint32_t getuint32(FILE *);
static uint16_t getuint16(FILE *);
static uint8_t getuint8(FILE *);
static void putuint32(uint32_t, FILE *);
static void putuint16(uint16_t, FILE *);
static void putuint8(uint8_t, FILE *);
static void error(const char *, ...);
static void fatal(const char *, ...);

static char *progname;

int
main(int argc, char *argv[])
{
	progname = argv[0];

	if (argc == 1)
		strip("a.out");
	else
		while (--argc)
			strip(*++argv);

	exit(EXIT_SUCCESS);
}

static void
strip(const char *filename)
{
	FILE *input, *output;
	char *tempfilename;
	char template[PATH_MAX];
	struct mnx_hdr mexec;
	struct stat statbuf;
	size_t symssize, segssize;
	uint8_t c;
	uint32_t relbase;

	if ((input = fopen(filename, "r")) == NULL) {
		error("can't open %s\n", filename);
		return;
	}

	mexec.a_magic[0] = getuint8(input);
	mexec.a_magic[1] = getuint8(input);
	mexec.a_flags = getuint8(input);
	mexec.a_cpu = getuint8(input);
	mexec.a_hdrlen = getuint8(input);
	mexec.a_unused = getuint8(input);
	mexec.a_version = getuint16(input);
	mexec.a_text = getuint32(input);
	mexec.a_data = getuint32(input);
	mexec.a_bss = getuint32(input);
	mexec.a_entry = getuint32(input);
	mexec.a_total = getuint32(input);
	mexec.a_syms = getuint32(input);

	if ((mexec.a_magic[0] != MNX_MAGIC0) ||
	    (mexec.a_magic[1] != MNX_MAGIC1) ||
	    (mexec.a_cpu != MNX_M68K)) {
		error("%s is not a MINIX-ST executable file\n", filename);
		fclose(input);
		return;
	}

	if (mexec.a_syms == 0) {
		fclose(input);
		return;
	}

	snprintf(template, PATH_MAX, "%s.strip.XXXXXX", filename);
	tempfilename = mktemp(template);
	if ((output = fopen(tempfilename, "w+")) == NULL) {
		error("can't create temporary file\n");
		fclose(input);
		return;
	}

	symssize = mexec.a_syms;
	segssize = mexec.a_text + mexec.a_data;

	mexec.a_syms = 0;

	putuint8(mexec.a_magic[0], output);
	putuint8(mexec.a_magic[1], output);
	putuint8(mexec.a_flags, output);
	putuint8(mexec.a_cpu, output);
	putuint8(mexec.a_hdrlen, output);
	putuint8(mexec.a_unused, output);
	putuint16(mexec.a_version, output);
	putuint32(mexec.a_text, output);
	putuint32(mexec.a_data, output);
	putuint32(mexec.a_bss, output);
	putuint32(mexec.a_entry, output);
	putuint32(mexec.a_total, output);
	putuint32(mexec.a_syms, output);

	while (segssize > 0) {
		c = getuint8(input);
		putuint8(c, output);
		segssize--;
	}

	if (fseek(input, symssize, SEEK_CUR) < 0) {
		error("can't skip symboltable in %s\n", filename);
		fclose(input);
		return;
	}

	relbase = getuint32(input);
	putuint32(relbase, output);
	if (relbase != 0) {
		do {
			c = getuint8(input);
			putuint8(c, output);
		} while (c != 0);
	}

	fclose(input);
	fclose(output);

	if (stat(filename, &statbuf) < 0) {
		error("can't stat %s\n", filename);
		return;
	}
	if (rename(tempfilename, filename) < 0) {
		error("can't rename %s to %s\n", tempfilename, filename);
		return;
	}
	if (chmod(filename, statbuf.st_mode) < 0) {
		error("can't chmod %s\n", filename);
		return;
	}

	return;
}

static uint32_t
getuint32(FILE *input)
{
	uint32_t result;

	result = getuint8(input);
	result = result << 8 | getuint8(input);
	result = result << 8 | getuint8(input);
	result = result << 8 | getuint8(input);

	return result;
}

static uint16_t
getuint16(FILE *input)
{
	uint16_t result;

	result = getuint8(input);
	result = result << 8 | getuint8(input);

	return result;
}

static uint8_t
getuint8(FILE *input)
{
	int result;

	if ((result = getc(input)) == EOF)
		fatal("read error\n");

	return result;
}

static void
putuint32(uint32_t value, FILE *output)
{
	putuint8(value >> 24, output);
	putuint8(value >> 16, output);
	putuint8(value >>  8, output);
	putuint8(value      , output);

	return;
}

static void
putuint16(uint16_t value, FILE *output)
{
	putuint8(value >>  8, output);
	putuint8(value      , output);

	return;
}

static void
putuint8(uint8_t value, FILE *output)
{
	if (putc(value, output) == EOF)
		fatal("write error\n");

	return;
}

static void
error(const char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);
	fprintf(stderr, "%s: ", progname);
	vfprintf(stderr, fmt, ap);
	va_end(ap);

	return;
}

static void
fatal(const char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);
	fprintf(stderr, "%s: ", progname);
	vfprintf(stderr, fmt, ap);
	va_end(ap);

	exit(EXIT_FAILURE);
}
